#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import numpy as np
from pandas import DataFrame, Series
from sklearn.model_selection import train_test_split, KFold


class Splitter(object):
    """
    Data splitter for train and test
    """

    def __init__(self,
                 sample_size,
                 *,
                 test_size=0.2,
                 cv=None,
                 random_state=None,
                 shuffle=True):
        """
        Parameters
        ----------
        sample_size: int
            Total sample size.
        test_size: float, int, None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. By default, the value is set to 0.2.
        cv: int, Series, list of string
            Number of folds. Must be at least 2. Default value is None.
        random_state: int, RandomState instance or None
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`. Default is None.
        shuffle: boolean
            Whether or not to shuffle the data before splitting.
        """
        self._sample_size = np.arange(sample_size)
        self._cv = cv
        self._random_state = random_state
        self._shuffle = shuffle
        self._test_size = test_size
        self._test = None
        self._train = None
        self.roll(
            test_size=test_size, random_state=random_state, shuffle=shuffle)

    def roll(self, *, sample_size=None, test_size=0.2, cv=None, random_state=None, shuffle=True):
        """
        Generate the indices of train-test data.

        Parameters
        ----------
        sample_size: int
            The row size of samples.
        test_size: float, int, None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. By default, the value is set to 0.2.
        cv: int or string
            Number of folds. Must be at least 2. Default value is None.
        random_state: int, RandomState instance or None
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`. Default is None.
        shuffle: boolean
            Whether or not to shuffle the data before splitting.
        """
        self._random_state = random_state
        self._cv = cv if cv is not None else self._cv
        self._sample_size = sample_size if sample_size is not None else self._sample_size
        self._test_size = test_size if test_size is not None else self._test_size
        self._shuffle = shuffle if shuffle is not None else self._shuffle
        self._train, self._test = train_test_split(
            self._sample_size,
            test_size=self._test_size,
            random_state=self._random_state,
            shuffle=self._shuffle)
        return self

    @property
    def size(self):
        return self._sample_size.size

    @staticmethod
    def _split(array, train, validate=None, test=None):
        def _size_check(array_):
            size = train.size
            if validate is not None:
                size = size + validate.size
            if test is not None:
                size = size + test.size
            if array_.shape[0] != size:
                raise ValueError(
                    'para `data` must have row size {} but got {}'.format(size, array_.shape[0]))

        # all to np.array
        if isinstance(array, list):
            array = np.array(array)
        if isinstance(array, (DataFrame, Series)):
            array = array.values
        if not isinstance(array, np.ndarray):
            raise TypeError(
                'must be a list or matrix-like object but got {}.'.format(array))

        # check size
        _size_check(array)

        # split data
        ret = [array[train]]
        if validate is not None:
            ret.append(array[validate])
        if test is not None:
            ret.append(array[test])
        return ret

        # raise error

    def cv(self, *arrays, less_for_train=False):
        """
        Split data with cross-validation.

        Parameters
        ----------
        *arrays: DataFrame, Series, ndarray, list
            Data for split. Must be a Sequence of indexables with same length / shape[0].
            If None, return the split indices.
        less_for_train: bool
            If true, use less data set for train.
            E.g. ``[1, 2, 3, 4, 5, 6, 7, 8, 9, 0]`` with 5 cv will be split into
            ``[1, 2]`` and ``[3, 4, 5, 6, 7, 8, 9, 0]``. Usually, ``[1, 2]`` (less one)
            will be used for validate. With ``less_for_train=True``, ``[1, 2]`` will be
            used for train. Default is ``False``.

        Yields
        -------
        list
            list containing split of inputs with cv. if inputs are None, only return
            the indices of split. if ``test_size`` is 0, test data/index will
            not return.
        """

        def group_cv():
            group = self._cv
            if isinstance(group, (list, np.ndarray)):
                group = Series(group)
            if not isinstance(group, Series):
                raise TypeError(
                    'must be a list or matrix-like object but got {}.'.format(group))
            group = group.rename('flag').reset_index(drop=True)
            group = group.iloc[self._train]

            for g in set(group):
                pure = group[group == g].index.values
                mixed = group[group != g].index.values
                if self._shuffle:
                    np.random.shuffle(pure)
                    np.random.shuffle(mixed)

                yield mixed, pure

            return

        if self._cv is None:
            raise ValueError('para `cv` must be set')

        if isinstance(self._cv, int):
            cv = KFold(n_splits=self._cv, shuffle=self._shuffle).split(self._train)
        else:
            cv = group_cv()
        for train, validate in cv:
            if less_for_train:
                train_, validate_, test_ = self._train[validate], self._train[train], self._test
            else:
                train_, validate_, test_ = self._train[train], self._train[validate], self._test
            if test_.size != 0:
                if len(arrays) == 0:
                    yield train_, validate_, test_
                else:
                    yield tuple([self._split(array, train_, validate_, test_) for array in arrays])
            else:
                if len(arrays) == 0:
                    yield train_, validate_
                else:
                    yield tuple([self._split(array, train_, validate_) for array in arrays])
        return

    def split(self, *arrays):
        """
        Split data.

        Parameters
        ----------
        *arrays: DataFrame, Series, ndarray, list
            Data for split. Must be a Sequence of indexables with same length / shape[0].
            If None, return the split indices.

        Returns
        -------
        tuple
            List containing split of inputs. if inputs are None, only return
            the indices of split. if ``test_size`` is 0, test data/index will
            not return.
        """
        if len(arrays) == 0:
            return self._train, self._test

        # if len(arrays) == 1:
        #     return _split(arrays[0])
        ret_ = [self._split(array, self._train, None, self._test) for array in arrays]
        ret__ = []
        for s in ret_:
            ret__ += s
        return tuple(ret__)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import re
from collections import defaultdict
from datetime import datetime as dt
from os import remove
from os.path import getmtime
from pathlib import Path
from shutil import rmtree
from warnings import warn

import pandas as pd
import requests
from ruamel.yaml import YAML
from sklearn.externals import joblib

from .._conf import __cfg_root__
from ..utils import Singleton
from ..utils import get_conf, get_dataset_url, get_data_loc, absolute_path, get_sha256


class LocalStorage(object):
    """
    Save data in a convenient way:

    .. code:: python

        import numpy as np
        np.random.seed(0)

        # some data
        data1 = np.random.randn(5, 5)
        data2 = np.random.randint(5, 5)

        # init Saver
        saver = LocalStorage('you_dataset_name')

        # save data
        saver(data1, data2)
        # or
        saver.save(data1, data2)

        # retriever data
        date = saver.last()  # last saved
        data = saver[0]  # by index
        for data in saver:  # as iterator
            do_something(data)

        # delete data
        save.rm(0)  # by index
        save.rm()  # delete 'you_dataset_name' dir

    See Also: :doc:`dataset`
    """

    def __init__(self, name=None, *, path=None, mk_dir=True, backend=None):
        """
        Parameters
        ----------
        name: str
            Name of dataset. Usually this is dir name contains data.
        path: str
            Absolute dir path.
        mk_dir: bool
            Ignore ``FileNotFoundError``.
        """
        self._backend = backend or joblib
        self._name = name
        if path is not None:
            self._path = Path(absolute_path(path, mk_dir)) / name
        else:
            self._path = Path(get_data_loc('userdata')) / name
        if not self._path.exists():
            if not mk_dir:
                raise FileNotFoundError()
            self._path.mkdir(parents=True)
        self._files = None
        self._make_file_index()

    @property
    def name(self):
        """
        Dataset name.

        Returns
        -------
        str
        """
        return self._name

    @property
    def path(self):
        """
        Dataset path

        Returns
        -------
        str
        """
        return str(self._path.parent)

    def _make_file_index(self):
        self._files = defaultdict(list)
        files = [f for f in self._path.iterdir() if f.match('*.pkl.*')]

        for f in files:
            # select data
            fn = '.'.join(f.name.split('.')[:-3])

            # for compatibility
            # fixme: will be removed at future
            if fn == 'unnamed':
                warn('file like `unnamed.@x` will be renamed to `@x`.', RuntimeWarning)
                new_name = '.'.join(f.name.split('.')[-3:])
                new_path = f.parent / new_name
                f.rename(new_path)
                f = new_path

            if fn == '':
                fn = 'unnamed'
            self._files[fn].append(f)

        for fs in self._files.values():
            if fs is not None:
                fs.sort(key=lambda f: getmtime(str(f)))

    def _load_data(self, file):
        if file.suffix == '.pd_':
            return pd.read_pickle(str(file))
        else:
            return self._backend.load(str(file))

    def _save_data(self, data, filename):
        self._path.mkdir(parents=True, exist_ok=True)
        if isinstance(data, pd.DataFrame):
            file = self._path / (filename + '.pkl.pd_')
            pd.to_pickle(data, str(file))
        else:
            file = self._path / (filename + '.pkl.z')
            self._backend.dump(data, str(file))

        return file

    def dump(self, fpath, *, rename=None, with_datetime=True):
        """
        Dump last checked dataset to file.

        Parameters
        ----------
        fpath: str
            Where save to.
        rename: str
            Rename pickle file. Omit to use dataset as name.
        with_datetime: bool
            Suffix file name with dumped time.

        Returns
        -------
        ret: str
            File path.
        """
        ret = {k: self._load_data(v[-1]) for k, v in self._files.items()}
        name = rename if rename else self._name
        if with_datetime:
            datetime = dt.now().strftime('-%Y-%m-%d_%H-%M-%S_%f')
        else:
            datetime = ''
        path_dir = Path(fpath).expanduser()
        if not path_dir.exists():
            path_dir.mkdir(parents=True, exist_ok=True)
        path = path_dir / (name + datetime + '.pkl.z')
        self._backend.dump(ret, str(path))

        return str(path)

    def last(self, name=None):
        """
        Return last saved data.

        Args
        ----
        name: str
            Data's name. Omit for access temp data

        Return
        -------
        ret:any python object
            Data stored in `*.pkl` file.
        """
        try:
            if name is None:
                return self._load_data(self._files['unnamed'][-1])
            if not isinstance(name, str):
                raise TypeError('para `name` must be string but got %s' % type(name))
            return self._load_data(self._files[name][-1])
        except IndexError:
            return None

    def rm(self, index, name=None):
        """
        Delete file(s) with given index.

        Parameters
        ----------
        index: int or slice
            Index of data. Data sorted by datetime.
        name: str
            Data's name. Omit for access unnamed data.
        """

        if not name:
            files = self._files['unnamed'][index]
            if not isinstance(files, list):
                remove(str(files))
            else:
                for f in files:
                    remove(str(f))
            del self._files['unnamed'][index]
            return

        files = self._files[name][index]
        if not isinstance(files, list):
            remove(str(files))
        else:
            for f in files:
                remove(str(f))
        del self._files[name][index]

    def clean(self, name=None):
        """
        Remove all data by name. Omit to remove hole dataset.

        Parameters
        ----------
        name: str
            Data's name.Omit to remove hole dataset.
        """
        if name is None:
            rmtree(str(self._path))
            self._files = list()
            self._files = defaultdict(list)
            return

        for f in self._files[name]:
            remove(str(f))
        del self._files[name]

    def __getattr__(self, name):
        """
        Returns sub-dataset.

        Parameters
        ----------
        name: str
            Dataset name.

        Returns
        -------
        self
        """
        sub_set = LocalStorage(name, path=str(self._path), backend=self._backend)
        setattr(self, name, sub_set)
        return sub_set

    def __repr__(self):
        cont_ls = ['<{}> include:'.format(self._name)]

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, len(v)))

        return '\n'.join(cont_ls)

    def __getitem__(self, item):

        # load file
        def _load_file(files, item_):
            _files = files[item_]
            if not isinstance(_files, list):
                return self._load_data(_files)
            return [self._load_data(f) for f in _files]

        if isinstance(item, tuple):
            try:
                key, index = item
            except ValueError:
                raise ValueError('except 2 parameters. [str, int or slice]')
            if not isinstance(key, str) or \
                    (not isinstance(index, int) and not isinstance(index, slice)):
                raise ValueError('except 2 parameters. [str, int or slice]')
            return _load_file(self._files[key], index)

        if isinstance(item, str):
            return self.__getitem__((item, slice(None, None, None)))

        return _load_file(self._files['unnamed'], item)

    def __call__(self, *unnamed_data, **named_data):
        """
        Same to self.save()
        """
        self.save(*unnamed_data, **named_data)

    def save(self, *unnamed_data, **named_data):
        """
        Save data with or without name.
        Data with same name will not be overwritten.

        Parameters
        ----------
        unnamed_data: any object
            Unnamed data.
        named_data: dict
            Named data as k,v pair.

        """

        def _get_file_index(fn):
            if len(self._files[fn]) != 0:
                return int(re.findall(r'@\d+\.', str(self._files[fn][-1]))[-1][1:-1])
            return 0

        num = 0
        for d in unnamed_data:
            if num == 0:
                num = _get_file_index('unnamed')
            num += 1
            f = self._save_data(d, '@' + str(num))
            self._files['unnamed'].append(f)

        for k, v in named_data.items():
            num = _get_file_index(k) + 1
            f = self._save_data(v, k + '.@' + str(num))
            self._files[k].append(f)


class Preset(Singleton):
    """
    Load data from embed dataset in XenonPy's or user create data saved in ``~/.xenonpy/cached`` dir.
    Also can fetch data by http request.

    This is sample to demonstration how to use is. Also see parameters documents for details.

    ::

        >>> from xenonpy.datatools import preset
        >>> elements = preset.load('elements')
        >>> elements.info()
        <class 'pandas.core.frame.DataFrame'>
        Index: 118 entries, H to Og
        Data columns (total 74 columns):
        atomic_number                    118 non-null int64
        atomic_radius                    88 non-null float64
        atomic_radius_rahm               96 non-null float64
        atomic_volume                    91 non-null float64
        atomic_weight                    118 non-null float64
        boiling_point                    96 non-null float64
        brinell_hardness                 59 non-null float64
        bulk_modulus                     69 non-null float64
        ...
    """

    # dataset = (
    #     'elements', 'elements_completed', 'mp_inorganic',
    #     'electron_density', 'sample_A', 'mp_structure'
    # )
    dataset = ('elements', 'elements_completed', 'mp_inorganic', 'mp_structure')
    # set to check params

    _dataset_dir = Path().home() / __cfg_root__ / 'dataset'
    _cached_dir = Path().home() / __cfg_root__ / 'cached'
    _yaml = YAML(typ='safe')
    _yaml.indent(mapping=2, sequence=4, offset=2)

    @classmethod
    def _http_data(cls, url, params=None, save_to=None, **kwargs):
        r = requests.get(url, params, **kwargs)
        r.raise_for_status()

        if 'filename' in r.headers:
            filename = r.headers['filename']
        else:
            filename = url.split('/')[-1]

        if save_to is None:
            save_to = str(cls._cached_dir / filename)
        with open(save_to, 'wb') as f:
            for chunk in r.iter_content(chunk_size=256 * 1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return save_to

    def __call__(self, data, *, sync=False):
        """
        Same to self.load()
        """
        return self.load(data=data, sync=sync)

    @classmethod
    def load(cls, data, *, sync=False):
        """
        load data.

        .. note::
            Try to load data from local at ``~/.xenonpy/dataset``.
            If no data, try to fetch them from remote repository.

        Args
        -----------
        data: str
            name of data.
        sync: bool
            If ``true``, try to sync data file with online version.

        Returns
        ------
        ret:DataFrame or Saver or local file path.
        """

        dataset = cls._dataset_dir / (data + '.pkl.pd_')
        sha256_file = cls._dataset_dir / 'sha256.yml'

        # fetch data from source if not in local
        if not dataset.exists():
            url = get_dataset_url(data)
            print('fetching dataset `{}` from {} because of none data file or outdated. '
                  'If you can\'t download it automatically, '
                  'just download it manually from the url above and '
                  'put it under `~/.xenonpy/dataset/`'.format(data, url))
            cls._http_data(url, save_to=str(dataset))

        # check sha256 value
        sha256_file.touch()  # make sure sha256_file file exist.
        sha256 = cls._yaml.load(sha256_file)
        if sha256 is None:
            sha256 = {}
        if data not in sha256:
            sha256_ = get_sha256(str(dataset))
            sha256[data] = sha256_
            cls._yaml.dump(sha256, sha256_file)
        else:
            sha256_ = sha256[data]

        if sha256_ != get_conf(data):
            warn(
                'local data {} is different from the repository version. '
                'use `load(data, sync=True)` to fix it.'.format(data), RuntimeWarning)
            if sync:
                url = get_dataset_url(data)
                print('fetching dataset `{}` from {} because of none data file or outdated. '
                      'you can download it manually from this url '
                      'then put it under `~/.xenonpy/dataset/`'.format(data, url))
                cls._http_data(url, save_to=str(dataset))
                sha256_ = get_sha256(str(dataset))
                sha256[data] = sha256_
                cls._yaml.dump(sha256, sha256_file)

        # return preset data
        return pd.read_pickle(str(dataset))

    @classmethod
    def _get_prop(cls, name):
        return cls.load(name)

    @property
    def elements(self):
        """
        Element properties from embed dataset.
        These properties are summarized from `mendeleev`_, `pymatgen`_, `CRC Handbook`_ and `magpie`_.

        See Also: :doc:`dataset`

        .. _mendeleev: https://mendeleev.readthedocs.io
        .. _pymatgen: http://pymatgen.org/
        .. _CRC Handbook: http://hbcponline.com/faces/contents/ContentsSearch.xhtml
        .. _magpie: https://bitbucket.org/wolverton/magpie

        Returns
        -------
        DataFrame:
            element properties in pd.DataFrame
        """
        return self._get_prop('elements')

    @property
    def mp_inorganic(self):
        """
        Inorganic properties summarized from `Materials Projects`_.

        .. _Materials Projects: https://www.materialsproject.org/

        Returns
        -------
        DataFrame:
            Properties in pd.DataFrame
        """
        return self._get_prop('mp_inorganic')

    @property
    def mp_structure(self):
        """
        Inorganic structures summarized from `Materials Projects`_.

        Returns
        -------
        DataFrame:
            Structures as dict that can be loaded by pymatgen.
        """
        return self._get_prop('mp_structure')

    @property
    def oqmd_inorganic(self):
        """
        Inorganic properties summarized from `OQMD`_.

        .. _OQMD: http://www.oqmd.org

        Returns
        -------
        DataFrame:
            Properties in pd.DataFrame
        """
        return self._get_prop('oqmd_inorganic')

    @property
    def oqmd_structure(self):
        """
        Inorganic structures summarized from `OQMD`_.

        Returns
        -------
        DataFrame:
            Structures as dict that can be loaded by pymatgen.
        """
        return self._get_prop('oqmd_structure')

    @property
    def elements_completed(self):
        """
        Completed element properties. [MICE]_ imputation used

        .. [MICE] `Int J Methods Psychiatr Res. 2011 Mar 1; 20(1): 40–49.`__
                    doi: `10.1002/mpr.329 <10.1002/mpr.329>`_

        .. __: https://www.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&retmode=ref&cmd=prlinks&id=21499542

        See Also: :doc:`dataset`

        Returns
        -------
            imputed element properties in pd.DataFrame
        """
        return self._get_prop('elements_completed')


preset = Preset()



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

__all__ = ['IQSPR', 'BayesianRidgeEstimator', 'NGram']

from .estimator import BayesianRidgeEstimator
from .iqspr import IQSPR
from .modifier import NGram



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import numpy as np
from pymatgen.core.composition import Composition as PMGComp

from .base import BaseFeaturizer, BaseDescriptor
from ..datatools.dataset import preset


class _CompositionalFeature(BaseFeaturizer):

    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """
        Base class for composition feature.
        """

        if include and exclude:
            raise ValueError(
                'Paratemer "include" and "exclude" are mutually exclusive.')
        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

        if elements is None:
            elements = preset.elements_completed
        if include is not None:
            elements = elements[include]
        if exclude is not None:
            elements = elements.drop(exclude, axis=1)

        self.elements = elements
        self.__authors__ = ['TsumiNa']

    def featurize(self, comp):
        elems_, nums_ = [], []
        if isinstance(comp, PMGComp):
            comp = comp.as_dict()
        for e, n in comp.items():
            elems_.append(e)
            nums_.append(n)
        return self._func(elems_, nums_)

    def _func(self, elems, nums):
        raise NotImplementedError

    @property
    def feature_labels(self):
        """
        Generate attribute names.
        Returns:
            ([str]) attribute labels.
        """
        raise NotImplementedError("feature_labels() is not defined!")


class WeightedAvgFeature(_CompositionalFeature):
    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(
            n_jobs=n_jobs, elements=elements, include=include, exclude=exclude, on_errors=on_errors,
            return_type=return_type)

    def _func(self, elems, nums):
        elems_ = self.elements.loc[elems, :]
        w_ = nums / np.sum(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['ave:' + s for s in self.elements]


class WeightedSumFeature(_CompositionalFeature):
    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(
            n_jobs=n_jobs, elements=elements, include=include, exclude=exclude, on_errors=on_errors,
            return_type=return_type)

    def _func(self, elems, nums):
        elems_ = self.elements.loc[elems, :]
        w_ = np.array(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['sum:' + s for s in self.elements]


class WeightedVarFeature(_CompositionalFeature):
    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(
            n_jobs=n_jobs, elements=elements, include=include, exclude=exclude, on_errors=on_errors,
            return_type=return_type)

    def _func(self, elems, nums):
        elems_ = self.elements.loc[elems, :]
        w_ = nums / np.sum(nums)
        mean_ = w_.dot(elems_)
        var_ = elems_ - mean_
        return w_.dot(var_ ** 2)

    @property
    def feature_labels(self):
        return ['var:' + s for s in self.elements]


class MaxFeature(_CompositionalFeature):
    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(
            n_jobs=n_jobs, elements=elements, include=include, exclude=exclude, on_errors=on_errors,
            return_type=return_type)

    def _func(self, elems, _):
        elems_ = self.elements.loc[elems, :]
        return elems_.max().values

    @property
    def feature_labels(self):
        return ['max:' + s for s in self.elements]


class MinFeature(_CompositionalFeature):
    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(
            n_jobs=n_jobs, elements=elements, include=include, exclude=exclude, on_errors=on_errors,
            return_type=return_type)

    def _func(self, elems, _):
        elems_ = self.elements.loc[elems, :]
        return elems_.min().values

    @property
    def feature_labels(self):
        return ['min:' + s for s in self.elements]


class Compositions(BaseDescriptor):
    """
    Calculate elemental descriptors from compound's composition.
    """

    def __init__(self, elements=None, *, n_jobs=-1, include=None,
                 exclude=None, on_errors='raise'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        include: list
            Column's names of elemental info that should be used in descriptor calculation.
        exclude: list
            Column's names of elemental info that should not be used in descriptor calculation.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        """

        super().__init__()
        self.n_jobs = n_jobs
        if elements is None:
            elements = preset.elements_completed
        if include is not None:
            elements = elements[include]
        if exclude is not None:
            elements = elements.drop(exclude, axis=1)
        self.elements = elements

        self.composition = WeightedAvgFeature(elements, n_jobs=n_jobs, on_errors=on_errors)
        self.composition = WeightedSumFeature(elements, n_jobs=n_jobs, on_errors=on_errors)
        self.composition = WeightedVarFeature(elements, n_jobs=n_jobs, on_errors=on_errors)
        self.composition = MaxFeature(elements, n_jobs=n_jobs, on_errors=on_errors)
        self.composition = MinFeature(elements, n_jobs=n_jobs, on_errors=on_errors)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from functools import partial
from pathlib import Path

import torch
from sklearn.externals import joblib

from ...datatools.dataset import LocalStorage
from ...utils import get_data_loc


class Checker(LocalStorage):
    """
    Check point.
    """

    class __SL(object):
        load = partial(torch.load, map_location=torch.device('cpu'))
        dump = torch.save

    def __init__(self, name, path=None, *, increment=True):
        """
        Parameters
        ----------
        name: str
            Model name.
        path: str
            Save path.
        """
        if not name:
            raise ValueError('need model name.')
        if path is None:
            path = get_data_loc('usermodel')

        if increment:
            i = 1
            while Path(path + '/' + name + '@' + str(i)).exists():
                i += 1
            _fpath = Path(path + '/' + name + '@' + str(i))
        else:
            _fpath = Path(path) / name
        self._name = _fpath.stem
        super().__init__(self._name, path=path)

    @classmethod
    def load(cls, model_path):
        p = Path(model_path)
        return cls(p.stem, str(p.parent), increment=False)

    @property
    def model_name(self):
        return self._name

    @property
    def describe(self):
        """
        Model's description.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        dict
            Description.
        """
        return self.last('describe')

    @property
    def init_model(self):
        """
        Last appended initial model.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        model:
            The last appended model.
        """
        self._backend = self.__SL
        try:
            return self.last('init_model')
        except FileNotFoundError:
            return None
        finally:
            self._backend = joblib

    @init_model.setter
    def init_model(self, model):
        """
        Set initial model.
        This action don't overwrite but add a new object.

        Parameters
        ----------
        model: torch.nn.Module
        """
        if isinstance(model, torch.nn.Module):
            self._backend = self.__SL
            super().__call__(init_model=model)
            self._backend = joblib
        else:
            raise TypeError(
                'except `torch.nn.Module` object but got {}'.format(
                    type(model)))

    @property
    def trained_model(self):
        """
        Last appended pre-trained model.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        model:
            The last appended model.
        """
        self._backend = self.__SL
        try:
            return self.last('trained_model')
        except FileNotFoundError:
            return None
        finally:
            self._backend = joblib

    @trained_model.setter
    def trained_model(self, model):
        """
        Set pre-trained model

        Parameters
        ----------
        model: torch.nn.Module
        """
        if isinstance(model, torch.nn.Module):
            self._backend = self.__SL
            super().__call__(trained_model=model)
            self._backend = joblib
        else:
            raise TypeError(
                'except `torch.nn.Module` object but got {}'.format(
                    type(model)))

    def __getitem__(self, item):
        if isinstance(item, int):
            self._backend = self.__SL
            try:
                cp = self.checkpoints[item]
                model_state = cp['model_state']
                del cp['model_state']
                return model_state, cp
            except IndexError:
                return None
            finally:
                self._backend = joblib
        raise TypeError('except int as checkpoint index but got {}'.format(
            type(item)))

    def __call__(self, **kwargs):
        self._backend = self.__SL
        self.checkpoints(kwargs)
        self._backend = joblib



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from copy import deepcopy
from types import MethodType

import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.base import BaseEstimator
from sklearn.linear_model import BayesianRidge

from ..base import BaseLogLikelihood
from ...descriptor.base import BaseDescriptor, BaseFeaturizer


class BayesianRidgeEstimator(BaseLogLikelihood):
    def __init__(self, descriptor, **estimators):
        """
        Bayesian Ridge Estimator.

        Parameters
        ----------
        descriptor: BaseFeaturizer or BaBaseDescriptorse
            Descriptor generator.
        estimators: BaseEstimator
            Bayesian estimators in scikit-learn style.
            When pass ``return_std=True`` to Estimator's ``predict`` method,
            ``y_std`` should be returned at second. e.g: ``BayesianRidge`` in scikit-learn.
        """
        if estimators:
            self._mdl = deepcopy(estimators)
        else:
            self._mdl = {}
        if not isinstance(descriptor, (BaseFeaturizer, BaseDescriptor)):
            raise TypeError('<descriptor> must be a subclass of <BaseFeaturizer> or <BaseDescriptor>')
        self._descriptor = descriptor

    @property
    def estimators(self):
        return self._mdl

    def __getitem__(self, item):
        return self._mdl[item]

    def __setitem__(self, key, value):
        if not (hasattr(value, 'predict') and isinstance(value.predict, MethodType)):
            raise TypeError('estimator must be a regressor in scikit-learn style')
        self._mdl[key] = deepcopy(value)

    # todo: implement scale function
    def fit(self, smiles, y=None, *, X_scaler=None, y_scaler=None, **kwargs):
        """
        Parameters
        ----------
        smiles: list of string
            SMILES for training.
        y: pd.DataFrame
            Target properties for training.
        X_scaler: Scaler (optional, not implement)
            Scaler for transform X.
        y_scaler: Scaler (optional, not implement)
            Scaler for transform y.
        kwargs: dict
            Parameters pass to BayesianRidge initialization.
        """

        if not isinstance(y, pd.DataFrame):
            raise TypeError('please package all properties into a pd.DataFrame')

        # remove NaN fromm X
        desc = self._descriptor.transform(smiles)
        desc = pd.DataFrame(data=desc).reset_index(drop=True)
        y = y.reset_index(drop=True)
        desc.dropna(inplace=True)
        y = y.loc[desc.index]

        for c in y:
            y_ = y[c]  # get target property.
            # remove NaN from y_
            y_.dropna(inplace=True)
            desc_ = desc.loc[y_.index]
            desc_ = desc_.values

            mdl = BayesianRidge(compute_score=True, **kwargs)
            mdl.fit(desc_, y_)
            self._mdl[c] = mdl

    def log_likelihood(self, smis, **targets):
        def _avoid_overflow(ll_):
            # log(exp(log(UP) - log(C)) - exp(log(LOW) - log(C))) + log(C)
            # where C = max(log(UP), max(LOW))
            ll_c = np.max(ll_)
            ll_ = np.log(np.exp(ll_[1] - ll_c) - np.exp(ll_[0] - ll_c)) + ll_c
            return ll_

        ll = np.repeat(-1000.0, len(smis))
        tar_fps = self._descriptor.transform(smis)
        tar_fps = pd.DataFrame(data=tar_fps).reset_index(drop=True)
        tmp = tar_fps.isna().any(axis=1)
        idx = [i for i in range(len(smis)) if ~tmp[i]]
        tar_fps.dropna(inplace=True)

        # calculate likelihood
        ll_mat = []
        for k, (low, up) in targets.items():  # k: target; v: (low, up)

            # predict mean, std for all smiles
            mean, std = self._mdl[k].predict(tar_fps, return_std=True)

            # calculate low likelihood
            low_ll = norm.logcdf(low, loc=np.asarray(mean), scale=np.asarray(std))

            # calculate up likelihood
            up_ll = norm.logcdf(up, loc=np.asarray(mean), scale=np.asarray(std))

            # zip low and up likelihood to a 1-dim array then save it.
            # like: [(tar_low_smi1, tar_up_smi1),  (tar_low_smi2, tar_up_smi2), ..., (tar_low_smiN, tar_up_smiN)]
            lls = zip(low_ll, up_ll)
            ll_mat.append(list(lls))

        # sum all ll along each smiles
        # ll_sum = [[sum_low_smi1, sum_up_smi1], [sum_low_smi2, sum_up_smi2],...,[sum_low_smiN, sum_up_smiN],]
        ll_sum = np.sum(np.array(ll_mat), axis=0)
        tmp = np.array([*map(_avoid_overflow, ll_sum)])

        np.put(ll, idx, tmp)
        return ll
